#lang racket

(define ALPHABET    ".ALPHABET")
(define STATES      ".STATES")
(define TRANSITIONS ".TRANSITIONS")
(define INPUT       ".INPUT")
(define EMPTY       ".EMPTY")

(define wlp4dfa (open-input-string
".ALPHABET
a-z 
A-Z 
0-9
(
)
{
}
=
!
<
>
+
-
*
/
%
,
;
[
]
&
.STATES
start!
LPAREN!
RPAREN!
LBRACE!
RBRACE!
BECOMES!
EQ!
EX
NE!
LT!
LE!
GT!
GE!
PLUS!
MINUS!
STAR!
SLASH!
PCT!
COMMA!
SEMI!
LBRACK!
RBRACK!
AMP!
ZERO!
NUM!
d!
e1!
l1!
e2!
t1!
DELETE!
ID!
e3!
l2!
s!
ELSE!
i1!
IF!
n1!
INT!
n2!
e4!
NEW!
N!
U!
L!
NULL!
p!
r1!
i2!
n3!
t2!
l3!
PRINTLN!
r2!
e5!
t3!
u!
r3!
RETURN!
w!
a!
i3!
WAIN!
h!
i4!
l4!
WHILE!
.TRANSITIONS
start   (                               LPAREN
start   )                               RPAREN
start   {                               LBRACE
start   }                               RBRACE
start   =                               BECOMES
BECOMES =                               EQ
start   !                               EX
EX      =                               NE
start   <                               LT
LT      =                               LE
start   >                               GT
GT      =                               GE
start   +                               PLUS
start   -                               MINUS
start   *                               STAR
start   /                               SLASH
start   %                               PCT
start   ,                               COMMA
start   ;                               SEMI
start   [                               LBRACK
start   ]                               RBRACK
start   &                               AMP
start   0                               ZERO
start   1-9                             NUM
NUM     0-9                             NUM
start   d                               d
d       e                               e1
d       a-d f-z A-Z 0-9                 ID
e1      l                               l1
e1      a-k m-z A-Z 0-9                 ID
l1      e                               e2
l1      a-d f-z A-Z 0-9                 ID
e2      t                               t1
e2      a-s u-z A-Z 0-9                 ID
t1      e                               DELETE
t1      a-d f-z A-Z 0-9                 ID
DELETE  a-z A-Z 0-9                     ID
ID      a-z A-Z 0-9                     ID
start   e                               e3
e3      l                               l2
e3      a-k m-z A-Z 0-9                 ID
l2      s                               s
l2      a-r t-z A-Z 0-9                 ID
s       e                               ELSE
s       a-d f-z A-Z 0-9                 ID
ELSE    a-z A-Z 0-9                     ID
start   i                               i1
i1      f                               IF
IF      a-z A-Z 0-9                     ID
i1      a-e g-m o-z A-Z 0-9             ID
i1      n                               n1
n1      t                               INT
INT     a-z A-Z 0-9                     ID
start   n                               n2
n2      e                               e4
n2      a-d f-z A-Z 0-9                 ID
e4      w                               NEW
e4      a-v x-z A-Z 0-9                 ID
NEW     a-z A-Z 0-9                     ID
start   N                               N
N       U                               U
N       a-z A-T V-Z 0-9                 ID
U       L                               L
U       a-z A-K M-Z 0-9                 ID
L       L                               NULL
L       a-z A-K M-Z 0-9                 ID
NULL    a-z A-Z 0-9                     ID
start   p                               p
p       r                               r1
p       a-q s-z A-Z 0-9                 ID
r1      i                               i2
r1      a-h j-z A-Z 0-9                 ID
i2      n                               n3
i2      a-m o-z A-Z 0-9                 ID
n3      t                               t2
n3      a-s u-z A-Z 0-9                 ID
t2      l                               l3
t2      a-k m-z A-Z 0-9                 ID
l3      n                               PRINTLN
PRINTLN a-z A-Z 0-9                     ID
l3      a-m o-z A-Z 0-9                 ID
start   r                               r2
r2      e                               e5
r2      a-d f-z A-Z 0-9                 ID
e5      t                               t3
e5      a-s u-z A-Z 0-9                 ID
t3      u                               u
t3      a-t v-z A-Z 0-9                 ID
u       r                               r3
u       a-q s-z A-Z 0-9                 ID
r3      n                               RETURN
RETURN  a-z A-Z 0-9                     ID
r3      a-m o-z A-Z 0-9                 ID
start   w                               w
w       a                               a
w       b-g i-z A-Z 0-9                 ID
a       i                               i3
a       a-h j-z A-Z 0-9                 i3
i3      n                               WAIN
i3      a-m o-z A-Z 0-9                 ID
WAIN    a-z A-Z 0-9                     ID
w       h                               h
h       i                               i4
h       a-h j-z A-Z 0-9                 ID
i4      l                               l4
i4      a-k m-z A-Z 0-9                 ID
l4      e                               WHILE
l4      a-d f-z A-Z 0-9                 ID
WHILE   a-z A-Z 0-9                     ID
start   a-c f-h j-m o q s-v x-z A-M O-Z ID
.INPUT
")
)

(define (cleanse s)
  (define len (string-length s))
  (define result s)
  (for ([i len])
    (define letter (string-ref s i))
    (cond[(and (< i (- len 1)) 
               (char=? letter #\/) 
               (char=? (string-ref s (+ i 1)) #\/))
           (set! result (substring s 0 i))])
  )
  result
)

(define (read-chars [in wlp4dfa])
  (define result
    (let helper ([accumulator empty])
      (define c (peek-char in))
      (cond [(eof-object? c)
             (cond [(empty? accumulator) eof]
                   [else accumulator])]
            [(char-whitespace? c)
             (cond [(empty? accumulator)
                    (read-char in)
                    (helper accumulator)]
                   [else accumulator])]
            [else (helper (cons (read-char in) accumulator))])))
  (cond [(eof-object? result) eof]
        [else (list->string (reverse result))]))

(define (range? s)
  (and (equal? (string-length s) 3) (equal? (string-ref s 1) #\-)))

(define (range->list s)
  (let collect-chars
    [(lo (char->integer (string-ref s 0)))
     (hi (char->integer (string-ref s 2)))]
    (cond [(<= lo hi)
           (cons (integer->char lo) (collect-chars (+ lo 1) hi))]
          [else empty])))

(define alphabet '())
(define states '())
(define transitions '())
(define input "")
(define-struct tok (kind lexeme) #:transparent)
(define-struct state (name initial accepting))
(define-struct transition (from-state to-state conditions))

; Locations in the program that you should modify to store the
; DFA information have been marked with double-semicolon comments:
;; (Double-semicolon comment)
(define (process-dfa [in wlp4dfa])
  (read-line in) ; Alphabet section (skip header)
  (let process-alphabet []
    (define s (read-chars in))
    (cond [(equal? s STATES) (void)]
          [(range? s)
           ;; Variable 's' contains a range of symbols
           ;; Can use range->list to turn it into a list of chars
           (set! alphabet (append alphabet (range->list s)))
           (process-alphabet)]
          [(equal? (string-length s) 1)
           (define c (string-ref s 0))
           ;; Variable 'c' contains a single symbol (a char)
           (set! alphabet (append alphabet (list c)))
           (process-alphabet)]))

  (read-line in) ; States section (skip header)
  (let process-states [(initial #t)]
    (define s (read-chars in))
    (cond [(equal? s TRANSITIONS) (void)]
          [else
           (define accepting
             (and (string-suffix? s "!") (> (string-length s) 1)))
           (define name (if accepting (string-trim s "!" #:left? #f) s))
           (define newstate (make-state name initial accepting))
           (set! states (cons newstate states))
           ;; Variable 'name' contains a state name
           ;; Variable 'initial' is #t if it's the initial state, #f otherwise
           ;; Variable 'accepting is #t if it's an accepting state, #f otherwise
           (process-states #f)]))

  (read-line in) ; Transitions section (skip header)
  (let process-transitions []
    (define line (read-line in))
    (cond [(equal? line INPUT) (void)]
          [else
           (define in (open-input-string line))
           (define from-state (read-chars in))
           (define reversed-rest
             (let collect [(accumulator empty)]
               (define s (read-chars in))
               (cond [(eof-object? s) accumulator]
                     [else (collect (cons s accumulator))])))
           ; reversed-rest has the elements in reverse order
           ; so the to-state is the first element
           (define to-state (first reversed-rest))
           (define symbols
             (append-map
              (lambda (s)
                (cond [(range? s) (range->list s)]
                      [(equal? (string-length s) 1) (list (string-ref s 0))]
                      [else empty]))
              (reverse (rest reversed-rest))))
           ;; DFA has transitions from 'from-state' to 'to-state'
           ;; on each symbol in the list 'symbols'
           (define newtransition (make-transition from-state to-state symbols))
           (set! transitions (cons newtransition transitions))
           (process-transitions)]))

  ; Input section (already skipped header)
  #|
  (let process-input []
    (define s (read-chars in))
    (cond [(eof-object? s) (void)]
          [else
           ;; Variable 's' contains an input string for the DFA
           (set! input (string-append input s))
           (process-input)]))
  |#
)

(process-dfa)

(define (process-input [in (current-input-port)])
  (define s (read-line in))
  (cond [(eof-object? s) (void)]
        [else
          (define cleaneds (cleanse s))
          (set! input (string-append input cleaneds))
          (process-input)
        ])
)

(process-input)

(define (isin? a lst)
  (define len (length lst))
  (define result #f)
  (for ([i len])
    (define c (list-ref lst i))
    (cond[(equal? c a) (set! result #t)])
  )
  result
)

(define (fetch s)
  (define len (length states))
  (define result 0)
  (for ([i len])
    (define state (list-ref states i))
    (define statename (state-name state))
    (cond[(string=? s statename) (set! result state)])
  )
  result
)

(define (move s letter)
  (define translen (length transitions))
  (define goto "NOT VALID")
  (define currentstate (state-name s))
  (for ([i translen])
    (define trans (list-ref transitions i))
    (define fromstate (transition-from-state trans))
    (define tostate (transition-to-state trans))
    (define conditionals (transition-conditions trans))
    (cond[(and (equal? currentstate fromstate) (isin? letter conditionals)) (set! goto tostate)])
  )
  (define result "NOT VALID")
  (cond[(not (string=? goto "NOT VALID")) (set! result (fetch goto))])
  result
)

(set! states (reverse states))
(set! transitions (reverse transitions))
(define result '())

(define (smm str)
    (define s (first states))
    (define token "")
    (define len (string-length str))
    (define valid #t)
    (for ([i len])
        (define nexti (+ i 1))
        (define letter (string-ref str i))
        (cond[(and (not (isin? letter alphabet)) 
                   (not (char-whitespace? letter))) (set! valid #f)])
        (cond[(equal? i (- len 1)) 
          (set! token (string-append token (string letter)))
          (set! s (move s letter))
          (cond[(and (not (equal? s "NOT VALID")) (state-accepting s))
            (define toadd (make-tok (state-name s) token))
            (set! result (cons toadd result))
            (set! token "")
            (set! s (first states))
          ]
          [else
            (set! valid #f)
          ])
        ]
        [(not (equal? (move s letter) "NOT VALID"))
          (set! token (string-append token (string letter)))
          (set! s (move s letter))
        ])
        (cond[(and (< i (- len 1)) (equal? (move s (string-ref str nexti)) "NOT VALID"))
                (cond[(state-accepting s)
                    (define toadd (make-tok (state-name s) token))
                    (set! result (cons toadd result))
                    (set! token "")
                    (set! s (first states))
                ]
                [else
                    (set! valid #f)
                ])
             ])
    )
    (cond[(not valid) (exit (eprintf "ERROR"))]
         [else
            (set! result (reverse result))
            (define resultlen (length result))
            (for ([j resultlen])
                (define kind (tok-kind (list-ref result j)))
                (define lexeme (tok-lexeme (list-ref result j)))
                (cond [(string=? kind "ZERO") (set! kind "NUM")]
                      [(or (string=? kind "d")
                           (string=? kind "e1")
                           (string=? kind "l1")
                           (string=? kind "e2")
                           (string=? kind "t1")
                           (string=? kind "e3")
                           (string=? kind "l2")
                           (string=? kind "s")
                           (string=? kind "i1")
                           (string=? kind "n1")
                           (string=? kind "n2")
                           (string=? kind "e4")
                           (string=? kind "N")
                           (string=? kind "U")
                           (string=? kind "L")
                           (string=? kind "p")
                           (string=? kind "r1")
                           (string=? kind "i2")
                           (string=? kind "n3")
                           (string=? kind "t2")
                           (string=? kind "l3")
                           (string=? kind "r2")
                           (string=? kind "e5")
                           (string=? kind "t3")
                           (string=? kind "u")
                           (string=? kind "r3")
                           (string=? kind "w")
                           (string=? kind "a")
                           (string=? kind "i3")
                           (string=? kind "h")
                           (string=? kind "i4")
                           (string=? kind "l4")) (set! kind "ID")])
                (cond[(not (string=? kind "start"))
                  (cond[(and (string=? kind "NUM") 
                             (> (string->number lexeme) 2147483647))
                             (exit (eprintf "ERROR"))])
                  (display kind)
                  (display " ")
                  (displayln lexeme)
                ])
            )
         ])
)
(smm input)